#define DONT_DEFINE_NEW
#include "HeapDebugger.h"
#include "CallStack.h"
#include <fstream>
#include <iostream>
#include <sstream>

static const size_t PAGE_SIZE = 4096;
const int DECOMMIT_TIME_TO_LIVE = 10;

HeapDebugger* HeapDebugger::instance = nullptr;

HeapDebugger::HeapDebugger(const char* filename) : prevHeapDebugger(instance),
                                                   filename(filename), currentAllocationId(0)
{
  //out = std::ofstream(filename);
  instance = this;
}

HeapDebugger::~HeapDebugger()
{
  WriteLeaks();
  if (instance == this)
  {
    instance = prevHeapDebugger;
  }
  else
  {
    throw new std::exception("If using multiple heap debuggers, they need to be destroyed in the opposite order from that which they were created.");
  }
}

HeapDebugger::AllocationData* HeapDebugger::PageAlignedAllocate(size_t size, bool isArray,
                                                                const char* file, 
                                                                unsigned int line)
{
  size_t numPages = size / PAGE_SIZE;
  if (size % PAGE_SIZE > 0)
  {
    ++numPages;
  }
  void* p = VirtualAlloc(0, PAGE_SIZE * (numPages + 1), MEM_RESERVE, PAGE_NOACCESS);

  if (p == nullptr)
  {
    return nullptr;
  }

  if (size > 0)
  {
    p = VirtualAlloc(p, PAGE_SIZE * numPages, MEM_COMMIT, PAGE_READWRITE);
  }

  PageData* page = (PageData*)malloc(sizeof(PageData));

  ++currentAllocationId;

  ReleaseOldPages();

  page->pageAddress = p;
  page->numPages = numPages;

  AllocationData* allocation = (AllocationData*)malloc(sizeof(AllocationData));

  new(&allocation->callStack) CallStack();
  allocation->addr = (unsigned char*)p + (PAGE_SIZE * numPages - size);
  allocation->size = size;
  STACK_TRACE(allocation->callStack);
  allocation->page = page;
  allocation->isArray = isArray;
  allocation->file = file;
  allocation->line = line;

  return allocation;
}


void HeapDebugger::DecommitPage(PageData* page)
{
  page->releaseAllocationId = currentAllocationId + DECOMMIT_TIME_TO_LIVE;
  VirtualFree(page->pageAddress, PAGE_SIZE * page->numPages, MEM_DECOMMIT);
  decomittedPages.push(page);
}

void* HeapDebugger::AllocateNoThrow(size_t size, bool isArray, const char* file, unsigned int line)
{
  AllocationData* allocation = PageAlignedAllocate(size, isArray, file, line);
  if (allocation == nullptr || allocation->addr == nullptr)
  {
    return nullptr;
  }
  else
  {
    allocationData.insert(std::make_pair((allocation->addr), allocation));
  }
  return allocation->addr;
}

void HeapDebugger::FreeNoThrow(void* ptr, bool isArray)
{
  if (ptr == nullptr)
  {
    return;
  }
  auto ptrIt = allocationData.find(ptr);
  if (ptrIt != allocationData.end())
  {
    if (ptrIt->second->isArray && !isArray)
    {
      std::basic_stringstream<char, std::char_traits<char>, Mallocator<char>> stream;
      stream << "Memory allocated with new[] and deleted with delete. Allocation at "
        << ptrIt->second->page << ":" << ptrIt->second->line
        << std::endl;
      stream << ptrIt->second->callStack;
      throw std::exception(stream.str().c_str());
    }
    else if (!ptrIt->second->isArray && isArray)
    {
      std::basic_stringstream<char, std::char_traits<char>, Mallocator<char>> stream;
      stream << "Memory allocated with new and deleted with delete[]. Allocation at "
        << ptrIt->second->page << ":" << ptrIt->second->line
        << std::endl;
      stream << ptrIt->second->callStack;
      throw std::exception(stream.str().c_str());
    }
    AllocationData* data = ptrIt->second;
    allocationData.erase(ptrIt);
    DecommitPage(data->page);
    free(data);
  }
  else
  {
    throw std::exception("This object was not allocated by new");
  }
}

void* HeapDebugger::Allocate(size_t size, bool isArray, const char* file, unsigned int line)
{
  void* allocated = AllocateNoThrow(size, isArray, file, line);
  if (allocated == nullptr)
  {
    throw std::bad_alloc();
  }
  return allocated;
}

void HeapDebugger::Free(void* ptr, bool isArray)
{
  FreeNoThrow(ptr, isArray);
}

void HeapDebugger::WriteLeaks()
{
  std::ostream* out;
  bool deleteOut;
  if (instance->filename == nullptr)
  {
    deleteOut = false;
    out = &std::cout;
  }
  else
  {
    deleteOut = true;
    //out = new std::ofstream(instance->filename, std::ios_base::out);
    out = (std::ostream*) malloc(sizeof(std::fstream));
    new(out)std::ofstream(instance->filename, std::ios_base::out);
  }

  for (auto data : allocationData)
  {
    *out << "Leaked " << data.second->size << " bytes at " << data.second->file << ":" 
      << data.second->line << std::endl;
    *out << data.second->callStack;
    *out << std::endl;
  }

  if (deleteOut)
  {
    free(out);
  }
}

HeapDebugger* HeapDebugger::GetInstance()
{
  return instance;
}

void HeapDebugger::ReleaseOldPages()
{
  while (!decomittedPages.empty() &&
         decomittedPages.front()->releaseAllocationId >= currentAllocationId)
  {
    PageData* page = decomittedPages.front();
    VirtualFree(page->pageAddress, PAGE_SIZE * (page->numPages + 1), MEM_RELEASE);
    decomittedPages.pop();
    free(page);
  }
}

void * operator new(size_t size)
{
  HeapDebugger* heap = HeapDebugger::GetInstance();
  if (heap == nullptr)
  {
    return malloc(size);
  }
  return heap->Allocate(size, false, "", 0);
}

void * operator new(size_t size, const std::nothrow_t&) throw ()
{
  HeapDebugger* heap = HeapDebugger::GetInstance();
  if (heap == nullptr)
  {
    return malloc(size);
  }
  return heap->AllocateNoThrow(size, false, "", 0);
}


void * operator new[](size_t size)
{
  HeapDebugger* heap = HeapDebugger::GetInstance();
  if (heap == nullptr)
  {
    return malloc(size);
  }
  return heap->Allocate(size, true, "", 0);
}

void * operator new[](size_t size, const std::nothrow_t&) throw ()
{
  HeapDebugger* heap = HeapDebugger::GetInstance();
  if (heap == nullptr)
  {
    return malloc(size);
  }
  return heap->AllocateNoThrow(size, true, "", 0);
}

void * operator new(size_t size, const char* file, unsigned int line)
{
  HeapDebugger* heap = HeapDebugger::GetInstance();
  if (heap == nullptr)
  {
    return malloc(size);
  }
  return heap->Allocate(size, false, file, line);
}

void * operator new(size_t size, const std::nothrow_t&, const char* file, unsigned int line) throw ()
{
  HeapDebugger* heap = HeapDebugger::GetInstance();
  if (heap == nullptr)
  {
    return malloc(size);
  }
  return heap->AllocateNoThrow(size, false, file, line);
}


void * operator new[](size_t size, const char* file, unsigned int line)
{
  HeapDebugger* heap = HeapDebugger::GetInstance();
  if (heap == nullptr)
  {
    return malloc(size);
  }
  return heap->Allocate(size, true, file, line);
}

void * operator new[](size_t size, const std::nothrow_t&, const char* file, unsigned int line) throw ()
{
  HeapDebugger* heap = HeapDebugger::GetInstance();
  if (heap == nullptr)
  {
    return malloc(size);
  }
  return heap->AllocateNoThrow(size, true, file, line);
}

void operator delete(void * ptr) throw ()
{
  HeapDebugger* heap = HeapDebugger::GetInstance();
  if (heap == nullptr)
  {
    return free(ptr);
  }
  heap->Free(ptr, false);
}

void operator delete(void * ptr, const std::nothrow_t&) throw ()
{
  HeapDebugger* heap = HeapDebugger::GetInstance();
  if (heap == nullptr)
  {
    return free(ptr);
  }
  heap->FreeNoThrow(ptr, false);
}


void operator delete[](void * ptr) throw ()
{
  HeapDebugger* heap = HeapDebugger::GetInstance();
  if (heap == nullptr)
  {
    return free(ptr);
  }
  heap->Free(ptr, true);
}

void operator delete[](void * ptr, const std::nothrow_t&) throw ()
{
  HeapDebugger* heap = HeapDebugger::GetInstance();
  if (heap == nullptr)
  {
    return free(ptr);
  }
  heap->FreeNoThrow(ptr, true);
}